/*************************************************************************
	Program	: Rotate 3D Part C2
	File	: rot3dc2.cpp
	Notes	: Set tab spaces to 4.
	
	Programmer	: James Johnson
	Date		: 1-21-94
	Version		: 1.0
	
	Comments	: Now lets optimize the rotation program by not using 
				  floating point calculations.  As you'll see, integer
				  calculation is much faster!
	
 ************************************************************************/

#include <stdio.h>
#include <math.h>

// Data type that I plan on using
//===================================================================
typedef struct{
	int x;
	int y;
	int z;
	unsigned Old_Offset;
}Point;

typedef struct{
	char r;
	char g;
	char b;
}Color;

// Function prototypes
//===================================================================
void setmode13(void);
inline void waitvrt(void);
void setmode03(void);
void inittbls(void);
void gotoxy(int x, int y);
char _kbhit(void);
char _getch(void);
void setcolor( unsigned char index, char r, char g, char b);
void setpalette(void);

// Global Constants
//===================================================================
const int Max_Points	= 125;
const int Max_Zc		= 1024;
const int Min_Zc		= 256;
const int Max_Index		= 360;
const int Def_Dx		= -1;
const int Def_Dy		= 2;
const int Def_Dz		= 0;
const int Xorg			= 160;
const int Yorg 			= 100;
const char ESC			= 0x1b;
const char SPC			= 0x20;

// Global Data
//===================================================================
Point Cube[Max_Points] = {	
	{-30, 30,-30, 0}, {-15, 30,-30, 0}, {  0, 30,-30, 0}, { 15, 30,-30, 0}, { 30, 30,-30, 0},
	{-30, 30,-15, 0}, {-15, 30,-15, 0}, {  0, 30,-15, 0}, { 15, 30,-15, 0}, { 30, 30,-15, 0},
	{-30, 30,  0, 0}, {-15, 30,  0, 0}, {  0, 30,  0, 0}, { 15, 30,  0, 0}, { 30, 30,  0, 0},
	{-30, 30, 15, 0}, {-15, 30, 15, 0}, {  0, 30, 15, 0}, { 15, 30, 15, 0}, { 30, 30, 15, 0},
	{-30, 30, 30, 0}, {-15, 30, 30, 0}, {  0, 30, 30, 0}, { 15, 30, 30, 0}, { 30, 30, 30, 0},
							                                                     
	{-30, 15,-30, 0}, {-15, 15,-30, 0}, {  0, 15,-30, 0}, { 15, 15,-30, 0}, { 30, 15,-30, 0},
	{-30, 15,-15, 0}, {-15, 15,-15, 0}, {  0, 15,-15, 0}, { 15, 15,-15, 0}, { 30, 15,-15, 0},
	{-30, 15,  0, 0}, {-15, 15,  0, 0}, {  0, 15,  0, 0}, { 15, 15,  0, 0}, { 30, 15,  0, 0},
	{-30, 15, 15, 0}, {-15, 15, 15, 0}, {  0, 15, 15, 0}, { 15, 15, 15, 0}, { 30, 15, 15, 0},
	{-30, 15, 30, 0}, {-15, 15, 30, 0}, {  0, 15, 30, 0}, { 15, 15, 30, 0}, { 30, 15, 30, 0},														                                                     
	
	{-30,  0,-30, 0}, {-15,  0,-30, 0}, {  0,  0,-30, 0}, { 15,  0,-30, 0}, { 30,  0,-30, 0},
	{-30,  0,-15, 0}, {-15,  0,-15, 0}, {  0,  0,-15, 0}, { 15,  0,-15, 0}, { 30,  0,-15, 0},
	{-30,  0,  0, 0}, {-15,  0,  0, 0}, {  0,  0,  0, 0}, { 15,  0,  0, 0}, { 30,  0,  0, 0},
	{-30,  0, 15, 0}, {-15,  0, 15, 0}, {  0,  0, 15, 0}, { 15,  0, 15, 0}, { 30,  0, 15, 0},
	{-30,  0, 30, 0}, {-15,  0, 30, 0}, {  0,  0, 30, 0}, { 15,  0, 30, 0}, { 30,  0, 30, 0},

	{-30,-15,-30, 0}, {-15,-15,-30, 0}, {  0,-15,-30, 0}, { 15,-15,-30, 0}, { 30,-15,-30, 0},
	{-30,-15,-15, 0}, {-15,-15,-15, 0}, {  0,-15,-15, 0}, { 15,-15,-15, 0}, { 30,-15,-15, 0},
	{-30,-15,  0, 0}, {-15,-15,  0, 0}, {  0,-15,  0, 0}, { 15,-15,  0, 0}, { 30,-15,  0, 0},
	{-30,-15, 15, 0}, {-15,-15, 15, 0}, {  0,-15, 15, 0}, { 15,-15, 15, 0}, { 30,-15, 15, 0},
	{-30,-15, 30, 0}, {-15,-15, 30, 0}, {  0,-15, 30, 0}, { 15,-15, 30, 0}, { 30,-15, 30, 0},

	{-30,-30,-30, 0}, {-15,-30,-30, 0}, {  0,-30,-30, 0}, { 15,-30,-30, 0}, { 30,-30,-30, 0},
	{-30,-30,-15, 0}, {-15,-30,-15, 0}, {  0,-30,-15, 0}, { 15,-30,-15, 0}, { 30,-30,-15, 0},
	{-30,-30,  0, 0}, {-15,-30,  0, 0}, {  0,-30,  0, 0}, { 15,-30,  0, 0}, { 30,-30,  0, 0},
	{-30,-30, 15, 0}, {-15,-30, 15, 0}, {  0,-30, 15, 0}, { 15,-30, 15, 0}, { 30,-30, 15, 0},
	{-30,-30, 30, 0}, {-15,-30, 30, 0}, {  0,-30, 30, 0}, { 15,-30, 30, 0}, { 30,-30, 30, 0}  };

int costbl[Max_Index];
int sintbl[Max_Index];

void main()
{   
	Point Disp[Max_Points];
	int adj1=0,adj2=0,adj3=1;
	int Xt, Yt, Zt;
	int Xc, Yc, Zc=Min_Zc;
	int i, Display = 1; 
	int Xang = Max_Index / 2, dx = Def_Dx;
	int Yang = 0, dy = Def_Dy;
	int Zang = 0, dz = Def_Dz;
	char quit = 0;
	unsigned char _far *vidmem = (unsigned char *)0xA0000000;

	setmode03();
	setcolor( 7, 0, 0, 0);
		
	printf("This program was created by Plexus ie niffty creature that clucks.\n");
	printf("It's the first little proggy that I've decided to release. I'd be \n");
	printf("very interested in any comments, or suggestions. Send email to    \n");
	printf("plexus@stein.u.washington.edu\n");
	printf("\n\tHere are some options:");
	printf("\n\t\t Esc \t\tTo Quit");
	printf("\n\t\tSpace\t\tTo Reset all values to orgional Values");
	printf("\n\t\t1 & 2\t\tIncreases & Decreases Delta X angles");
	printf("\n\t\t3 & 4\t\tIncreases & Decreases Delta Y angles");
	printf("\n\t\t5 & 6\t\tIncreases & Decreases Delta Z angles");
	printf("\n\t\t  7  \t\tToggle Wait VRT in area 1");
	printf("\n\t\t  8  \t\tToggle Wait VRT in area 2");
	printf("\n\t\t  9  \t\tToggle Wait VRT in area 3");
	printf("\n\t\t  D  \t\tToggle the display of angles");
	printf("\n\t\t  +  \t\tIncrease the Distance ie push object away");
	printf("\n\t\t  -  \t\tDecrease the Distance ie pull object closer");
	printf("\n\nPress any key to begin the trip...");
	
	for(i=0; i < 64; i++){
		waitvrt();
		setcolor(0,i,i,i);
		waitvrt();
		setcolor(7,i,i,i);
	}
	
	for(i=63; i >= 0; i--){
		waitvrt();
		setcolor(0,i,i,i);
	}

	_getch();
			
	for(i = 63; i > 0; i--){
		waitvrt();
		setcolor(7,i,i,i);
	}

	inittbls();
	setmode13();
	setpalette();
	
	while(!quit){
	    
	    // Precalculation
	    Xc = 50 * sintbl[Xang];
	    Xc = (Xc >> 8) + Xorg;
	    Yc = 30 * costbl[Yang];
	    Yc = (Yc >> 8) + Yorg;

		if(adj1) waitvrt();	    
		for(i=0; i < Max_Points; i++)
			*(vidmem + Disp[i].Old_Offset) = 0;

		// Rotate the Object.
		for(i=0; i < Max_Points; i++){
			// Rotate X axis
			Yt = Cube[i].y * costbl[Xang] - Cube[i].z * sintbl[Xang];
			Zt = Cube[i].y * sintbl[Xang] + Cube[i].z * costbl[Xang];
			Yt = Yt >> 8;
			Zt = Zt >> 8;
			
			// Rotate Y axis
			Xt = Cube[i].x * costbl[Yang] - Zt * sintbl[Yang];
			Zt = Cube[i].x * sintbl[Yang] + Zt * costbl[Yang];
			Xt = Xt >> 8;
			Disp[i].z = Zt >> 8;
			
			// Rotate Z axis. We must not change the values of Xt or Yt
			// before the calculation is done.
			Disp[i].x = Xt * costbl[Zang] - Yt * sintbl[Zang];
			Disp[i].y = Xt * sintbl[Zang] + Yt * costbl[Zang];

			// Add psudeo perspective and center it on the screen.
			// Notice that (Disp[i].z + Zc) > 0
			// Added info:
			// The smaller A & B are the weirder this will look, it'll
			// be like approaching a huge building. Actually A is muled
			// by 256.
			//			 A								B
			//			 |								|
			//			 V								V
			Disp[i].x = Disp[i].x / (Disp[i].z + Zc) + Xc;
			Disp[i].y = Disp[i].y / (Disp[i].z + Zc) + Yc;
		}

		if(adj2) waitvrt();		
		
		for(i=0; i < Max_Points; i++){
			Disp[i].Old_Offset = Disp[i].x + Disp[i].y * 320;
			*(vidmem + Disp[i].Old_Offset) = ((Disp[i].z + 30) % 80) + 64;
		}
		
		if(adj3) waitvrt();
		
		if(Display){
			gotoxy( 0,0);
			printf("Xang = %3d    Yang = %3d    Zang = %3d", Xang, Yang, Zang);		
		}
			
		if(_kbhit())
			switch(_getch()){
				case ESC	:	quit = 1;
						break;
				case SPC	:	Xang = Max_Index / 2;
								Yang = 0;
								Zang = 0;
								Zc = Min_Zc;
								dx = Def_Dx;
								dy = Def_Dy;
								dz = Def_Dz;
								adj1 = 0;
								adj2 = 0;
								adj3 = 1;
						break;
				case 'D'	:
				case 'd'	:	Display ^= 1;
								gotoxy(0,0);
								printf("                                          ");
						break;
				case '1'	:	dx++;
						break;
				case '2'	:	dx--;
						break;
				case '3'	:	dy++;
						break;
				case '4'	:	dy--;
						break;
				case '5'	:	dz++;
						break;
				case '6'	:	dz--;
						break;
				case '+'	:	if(Zc < (Max_Zc - 8)) Zc += 8;
						break;
				case '-'	:	if(Zc > (Min_Zc + 8)) Zc -= 8;
						break;
				case '7'	:	adj1 ^= 1;
						break;
				case '8'	:	adj2 ^= 1;
						break;
				case '9'	:	adj3 ^= 1;
						break;
			}

			
		Xang += dx;
		Yang += dy;
		Zang += dz;
		
		if(Xang < 0)
			Xang = Max_Index + ((Xang < 0) ? Xang: -Xang);
		if(Xang >= Max_Index)
			Xang = Xang - Max_Index;
			
		if(Yang < 0)
			Yang = Max_Index + ((Yang < 0) ? Yang: -Yang);
		if(Yang >= Max_Index)
			Yang = Yang - Max_Index;
			
		if(Zang < 0)
			Zang = Max_Index + ((Zang < 0) ? Zang: -Zang);
		if(Zang >= Max_Index)
			Zang = Zang - Max_Index;
	}
	setmode03();
	
	setcolor(7,0,0,0);
	
	printf("\nThis little proggy was brought to you by James Johnson.");
	printf("\nWritten in C/C++ with some assembly.");
	                
	for(i = 0; i < 128; i++){
		waitvrt();
		setcolor(7,i/2,i/2,i/2);
	}

	for(i = 127; i > 0; i--){
		waitvrt();
		setcolor(7,i/2,i/2,i/2);
	}
	setmode03();
}
			
		
void setmode13(void)
{
	__asm{
			mov		ax,13h
			int 	10h
	}
}

inline void waitvrt(void)
{
	__asm{
			mov		dx,3dah
	VRT:
			in		al,dx
			test	al,8
			jz		VRT
	NoVRT:
			in		al,dx
			test	al,8
			jnz		NoVRT
	}
}

void inittbls(void)
{
	double PI = 3.141592654;
	int i;
	for(i = 0; i < Max_Index; i++){
		costbl[i] = (int)(256 * cos(i * PI / 180) + .5);
		sintbl[i] = (int)(256 * sin(i * PI / 180) + .5);
	}
}

void setmode03(void)
{
	__asm{
			mov		ax,03
			int		10h
	}
}

void setcolor( unsigned char index, char r, char g, char b)
{
	__asm{
			mov		dx,0x3c8
			mov		al,index
			out		dx,al
			inc		dx
			mov		al,r
			out		dx,al
			mov		al,g
			out		dx,al
			mov		al,b
			out		dx,al
	}
}
			

void setpalette(void)
{
	int i;
	
	for(i=0; i < 64; i++)
		setcolor(i+64, 63-i, 63-i, 63-i);
}
	
void gotoxy(int x, int y)
{
	__asm{
			mov		ah,02
			mov		dl,byte ptr x
			mov		dh,byte ptr y
			mov		bh,0
			int		10h
	}
}	
		
char _kbhit(void)
{
	char chk=0;
	
	__asm{
			mov		ax,0100h
			int		16h
			jz		END_ASM
			mov		chk,1
		END_ASM:	
	}
	return chk;
}

char _getch(void)
{
	char chr;
	
	__asm{
			mov		ah,0h
			int		16h
			mov		chr,al
	}
	return chr;
}